# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from flask import Flask, request

from swh.scanner.exceptions import LargePayloadExc

from .data import present_swhids


def create_app():
    app = Flask(__name__)

    @app.route("/")
    def index():
        return "SWH scanner API"

    @app.route("/known/", methods=["POST"])
    def known():
        swhids = request.get_json()

        if len(swhids) > 900:
            raise LargePayloadExc(
                "The maximum number of SWHIDs this endpoint can receive is 900"
            )

        res = {swhid: {"known": False} for swhid in swhids}
        for swhid in swhids:
            if swhid in present_swhids:
                res[swhid]["known"] = True

        return res

    return app



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dash_html_components as html

from swh.scanner.dashboard.dashboard import generate_table_body


def test_generate_table_body(example_tree, temp_folder):
    subdir_path = temp_folder["subdir"]

    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True)

    generated_body = generate_table_body(subdir_path, example_tree)

    expected_body = [
        html.Tbody(
            [
                html.Tr(
                    [
                        html.Td("✔"),
                        html.Td(
                            html.A(
                                children="filesample.txt",
                                href=f"file://{subdir_path}/filesample.txt",
                            )
                        ),
                        html.Td("swh:1:cnt:e69de29bb2d1d6434b8b29ae775ad8c2e48c5391"),
                    ]
                ),
                html.Tr(
                    [
                        html.Td("✔"),
                        html.Td(
                            html.A(
                                children="filesample2.txt",
                                href=f"file://{subdir_path}/filesample2.txt",
                            )
                        ),
                        html.Td("swh:1:cnt:e69de29bb2d1d6434b8b29ae775ad8c2e48c5391"),
                    ]
                ),
            ]
        )
    ]

    # workaround: dash_html_component.__eq__ checks for object identity only
    assert str(generated_body) == str(expected_body)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

correct_api_response = {
    "swh:1:dir:17d207da3804cc60a77cba58e76c3b2f767cb112": {"known": False},
    "swh:1:dir:01fa282bb80be5907505d44b4692d3fa40fad140": {"known": True},
    "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904": {"known": True},
}

# present SWHIDs inside /data/sample-folder
present_swhids = [
    "swh:1:cnt:7c4c57ba9ff496ad179b8f65b1d286edbda34c9a",  # quotes.md
    "swh:1:cnt:68769579c3eaadbe555379b9c3538e6628bae1eb",  # some-binary
    "swh:1:dir:9619a28687b2462efbb5be816bc1185b95753d93",  # barfoo2/
    "swh:1:dir:07d4d9ec5c406632d203dbd4631e7863612a0326",  # toexclude/
]


to_exclude_swhid = "swh:1:dir:07d4d9ec5c406632d203dbd4631e7863612a0326"



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

import dash
from dash.dependencies import Input, Output
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objects as go

from ..model import Tree


def generate_table_body(dir_path: Path, source: Tree):
    """
    Generate the data_table from the path taken from the chart.

    For each file builds the html table rows showing the known status, a local link to
    the file and the relative SoftWare Heritage persistent IDentifier (SWHID).
    """
    data = []
    for file_info in source.get_files_from_dir(dir_path):
        for file_path, attr in file_info.items():
            file_path = Path(file_path)
            file_name = file_path.parts[len(file_path.parts) - 1]
            data.append(
                html.Tr(
                    [
                        html.Td("✔" if attr["known"] else ""),
                        html.Td(
                            html.A(file_name, href="file://" + str(file_path.resolve()))
                        ),
                        html.Td(attr["swhid"]),
                    ]
                )
            )

    return [html.Tbody(data)]


def run_app(graph_obj: go, source: Tree):
    app = dash.Dash(__name__)
    fig = go.Figure().add_trace(graph_obj)

    fig.update_layout(height=800,)

    table_header = [
        html.Thead(html.Tr([html.Th("KNOWN"), html.Th("FILE NAME"), html.Th("SWHID")]))
    ]

    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Div(
                        [dcc.Graph(id="sunburst_chart", figure=fig),], className="col",
                    ),
                    html.Div(
                        [
                            html.H3(id="directory_title"),
                            dbc.Table(
                                id="files_table",
                                hover=True,
                                responsive=True,
                                striped=True,
                            ),
                        ],
                        className="col",
                    ),
                ],
                className="row",
            ),
        ]
    )

    @app.callback(
        [Output("files_table", "children"), Output("directory_title", "children")],
        [Input("sunburst_chart", "clickData")],
    )
    def update_files_table(click_data):
        """
        Callback that takes the input (directory path) from the chart and
        update the `files_table` children with the relative files.

        """
        if click_data is not None:
            raw_path = click_data["points"][0]["label"]
            full_path = (
                source.path.joinpath(raw_path)
                if raw_path != str(source.path)
                else Path(raw_path)
            )
            return table_header + generate_table_body(full_path, source), str(full_path)
        else:
            return "", ""

    app.run_server(debug=True, use_reloader=True)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scanner.exceptions import InvalidDirectoryPath


def test_tree_add_node(example_tree, temp_folder):
    avail_paths = temp_folder["paths"].keys()

    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, False)

    for path, node in example_tree.children.items():
        assert path in avail_paths
        if node.children:
            for subpath, subnode in node.children.items():
                assert subpath in avail_paths


def test_to_json_no_one_present(example_tree, temp_folder):
    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, False)

    result = example_tree.to_dict()

    assert len(result) == 6

    for _, node_info in result.items():
        assert node_info["known"] is False


def test_get_json_tree_all_present(example_tree, temp_folder):
    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True)

    result = example_tree.to_dict()

    assert len(result) == 6

    for _, node_info in result.items():
        assert node_info["known"] is True


def test_get_json_tree_only_one_present(example_tree, temp_folder):
    filesample_path = temp_folder["filesample"]

    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True if path == filesample_path else False)

    result = example_tree.to_dict()

    assert len(result) == 6

    for path, node_attr in result.items():
        if path == "subdir/filesample.txt":
            assert node_attr["known"] is True
        else:
            assert node_attr["known"] is False


def test_get_directories_info(example_tree, temp_folder):
    root_path = temp_folder["root"]
    filesample_path = temp_folder["filesample"]
    filesample2_path = temp_folder["filesample2"]
    subdir_path = temp_folder["subdir"].relative_to(root_path)
    subsubdir_path = temp_folder["subsubdir"].relative_to(root_path)

    for path, swhid in temp_folder["paths"].items():
        if path == filesample_path or path == filesample2_path:
            example_tree.add_node(path, swhid, True)
        else:
            example_tree.add_node(path, swhid, False)

    directories = example_tree.get_directories_info(example_tree.path)

    assert subsubdir_path not in directories
    assert directories[subdir_path] == (2, 2)


def test_get_files_from_dir(example_tree, temp_folder):
    subdir_path = temp_folder["subdir"]

    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True)

    files = example_tree.get_files_from_dir(subdir_path)
    assert len(files) == 2


def test_get_files_source_path(example_tree, temp_folder):
    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True)

    files = example_tree.get_files_from_dir(example_tree.path)
    assert len(files) == 1


def test_get_files_from_dir_raise_exception(example_tree, temp_folder):
    for path, swhid in temp_folder["paths"].items():
        example_tree.add_node(path, swhid, True)

    with pytest.raises(InvalidDirectoryPath):
        example_tree.get_files_from_dir("test/")



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
The purpose of this module is to display and to interact with the result of the
scanner contained in the model.

The `sunburst` function generates a navigable sunburst chart from the
directories information retrieved from the model. The chart displays for
each directory the total number of files and the percentage of file known.

The size of the directory is defined by the total number of contents whereas
the color gradient is generated relying on the percentage of contents known.
"""

from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.offline import offline


def build_hierarchical_df(
    dirs_dataframe: pd.DataFrame,
    levels: List[str],
    metrics_columns: List[str],
    root_name: str,
) -> pd.DataFrame:
    """
        Build a hierarchy of levels for Sunburst or Treemap charts.

        For each directory the new dataframe will have the following
        information:

        id: the directory name
        parent: the parent directory of id
        contents: the total number of contents of the directory id and
        the relative subdirectories
        known: the percentage of contents known relative to computed
        'contents'

        Example:
        Given the following dataframe:

        .. code-block:: none

            lev0     lev1                contents  known
             ''       ''                 20        2     //root
            kernel   kernel/subdirker    5         0
            telnet   telnet/subdirtel    10        4

        The output hierarchical dataframe will be like the following:

        .. code-block:: none

              id                parent    contents  known
                                          20        10.00
           kernel/subdirker     kernel    5         0.00
           telnet/subdirtel     telnet    10        40.00
                                total     20        10.00
           kernel               total     5         0.00
           telnet               total     10        40.00
           total                          35        17.14

        To create the hierarchical dataframe we need to iterate through
        the dataframe given in input relying on the number of levels.

        Based on the previous example we have to do two iterations:

        iteration 1
        The generated dataframe 'df_tree' will be:

        .. code-block:: none

            id                parent   contents  known
                                       20        10.0
            kernel/subdirker  kernel   5         0.0
            telnet/subdirtel  telnet   10        40.0

        iteration 2
        The generated dataframe 'df_tree' will be:

        .. code-block:: none

            id       parent   contents  known
                     total    20        10.0
            kernel   total    5         0.0
            telnet   total    10        40.0

        Note that since we have reached the last level, the parent given
        to the directory id is the directory root.

        The 'total' row il computed by adding the number of contents of the
        dataframe given in input and the average of the contents known on
        the total number of contents.

    """

    def compute_known_percentage(contents: pd.Series, known: pd.Series) -> pd.Series:
        """This function compute the percentage of known contents and generate
           the new known column with the percentage values.

           It also assures that if there is no contents inside a directory
           the percentage is zero

        """
        known_values = []
        for idx, content_val in enumerate(contents):
            if content_val == 0:
                known_values.append(0)
            else:
                percentage = known[idx] / contents[idx] * 100
                known_values.append(percentage)

        return pd.Series(np.array(known_values))

    complete_df = pd.DataFrame(columns=["id", "parent", "contents", "known"])
    # revert the level order to start from the deepest
    levels = [level for level in reversed(levels)]
    contents_col = metrics_columns[0]
    known_col = metrics_columns[1]

    df_tree_list = []
    for i, level in enumerate(levels):
        df_tree = pd.DataFrame(columns=["id", "parent", "contents", "known"])
        dfg = dirs_dataframe.groupby(levels[i:]).sum()
        dfg = dfg.reset_index()
        df_tree["id"] = dfg[level].copy()
        if i < len(levels) - 1:
            # copy the parent directories (one level above)
            df_tree["parent"] = dfg[levels[i + 1]].copy()
        else:
            # last level reached
            df_tree["parent"] = root_name

        # copy the contents column
        df_tree["contents"] = dfg[contents_col]
        # compute the percentage relative to the contents
        df_tree["known"] = compute_known_percentage(dfg[contents_col], dfg[known_col])

        df_tree_list.append(df_tree)

    complete_df = complete_df.append(df_tree_list, ignore_index=True)

    # create the main parent
    total_contents = dirs_dataframe[contents_col].sum()
    total_known = dirs_dataframe[known_col].sum()
    total_avg = total_known / total_contents * 100

    total = pd.Series(
        dict(id=root_name, parent="", contents=total_contents, known=total_avg)
    )

    complete_df = complete_df.append(total, ignore_index=True)

    return complete_df


def compute_max_depth(dirs_path: List[Path], root: Path) -> int:
    """Compute the maximum depth level of the given directory paths.

       Example: for `var/log/kernel/` the depth level is 3

    """
    max_depth = 0
    for dir_path in dirs_path:
        if dir_path == root:
            continue

        dir_depth = len(dir_path.parts)
        if dir_depth > max_depth:
            max_depth = dir_depth

    return max_depth


def generate_df_from_dirs(
    dirs: Dict[Path, Tuple[int, int]], columns: List[str], root: Path, max_depth: int,
) -> pd.DataFrame:
    """Generate a dataframe from the directories given in input.

    Example:
    given the following directories as input

    .. code-block:: python

        dirs = {
            '/var/log/': (23, 2),
            '/var/log/kernel': (5, 0),
            '/var/log/telnet': (10, 3)
        }

    The generated dataframe will be:

    .. code-block:: none

        lev0   lev1       lev2             contents  known
        'var'  'var/log'   ''              23        2
        'var'  'var/log' 'var/log/kernel'  5         0
        'var'  'var/log' 'var/log/telnet'  10        3

    """

    def get_parents(path: Path):
        parts = path.parts[1:] if path.parts[0] == "/" else path.parts

        for i in range(1, len(parts) + 1):
            yield "/".join(parts[0:i])

    def get_dirs_array():
        for dir_path, contents_info in dirs.items():
            empty_lvl = max_depth - len(dir_path.parts)

            if dir_path == root:
                # ignore the root but store contents information
                yield [""] * (max_depth) + list(contents_info)
            else:
                yield list(get_parents(dir_path)) + [""] * empty_lvl + list(
                    contents_info
                )

    df = pd.DataFrame(
        np.array([dir_array for dir_array in get_dirs_array()]), columns=columns
    )

    df["contents"] = pd.to_numeric(df["contents"])
    df["known"] = pd.to_numeric(df["known"])

    return df


def generate_sunburst(
    directories: Dict[Path, Tuple[int, int]], root: Path
) -> go.Sunburst:
    """Generate a sunburst chart from the directories given in input.

    """
    max_depth = compute_max_depth(list(directories.keys()), root)
    metrics_columns = ["contents", "known"]
    levels_columns = ["lev" + str(i) for i in range(max_depth)]

    df_columns = levels_columns + metrics_columns
    dirs_df = generate_df_from_dirs(directories, df_columns, root, max_depth)

    hierarchical_df = build_hierarchical_df(
        dirs_df, levels_columns, metrics_columns, str(root)
    )

    sunburst = go.Sunburst(
        labels=hierarchical_df["id"],
        parents=hierarchical_df["parent"],
        values=hierarchical_df["contents"],
        branchvalues="total",
        marker=dict(
            colors=hierarchical_df["known"],
            colorscale="matter",
            cmid=50,
            showscale=True,
        ),
        hovertemplate="""<b>%{label}</b>
            <br>Files: %{value}
            <br>Known: <b>%{color:.2f}%</b>""",
        name="",
    )

    return sunburst


def offline_plot(graph_object: go):
    """Plot a graph object to an html file
    """
    fig = go.Figure()
    fig.add_trace(graph_object)
    offline.plot(fig, filename="chart.html")



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.scanner.db import Db

from .data import present_swhids

CHUNK_SIZE = 1000


def test_db_create_from(tmp_path, test_swhids_sample):
    tmp_dbfile = tmp_path / "tmp_db.sqlite"

    db = Db(tmp_dbfile)
    cur = db.conn.cursor()
    db.create_from(test_swhids_sample, CHUNK_SIZE, cur)

    for swhid in present_swhids:
        cur = db.conn.cursor()
        assert db.known(swhid, cur)


def test_db_create_from_one_not_present(tmp_path, test_swhids_sample):
    not_present_swhid = "swh:1:cnt:fa8eacf43d8646129ae8adfa1648f9307d999999"
    swhids = present_swhids + [not_present_swhid]

    tmp_dbfile = tmp_path / "tmp_db.sqlite"

    db = Db(tmp_dbfile)
    cur = db.conn.cursor()
    db.create_from(test_swhids_sample, CHUNK_SIZE, cur)

    for swhid in swhids:
        cur = db.conn.cursor()
        if swhid != not_present_swhid:
            assert db.known(swhid, cur)
        else:
            assert not db.known(swhid, cur)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import os
from pathlib import Path
import shutil

import aiohttp
from aioresponses import aioresponses  # type: ignore
import pytest

from swh.model.cli import swhid_of_dir, swhid_of_file
from swh.scanner.model import Tree

from .data import present_swhids
from .flask_api import create_app


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.fixture
def event_loop():
    """Fixture that generate an asyncio event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


@pytest.fixture
async def aiosession():
    """Fixture that generate an aiohttp Client Session."""
    session = aiohttp.ClientSession()
    yield session
    session.detach()


@pytest.fixture(scope="function")
def temp_folder(tmp_path):
    """Fixture that generates a temporary folder with the following
    structure:

    .. code-block:: python

        root = {
            subdir: {
                subsubdir
                filesample.txt
                filesample2.txt
            }
            subdir2
            subfile.txt
        }
    """
    root = tmp_path
    subdir = root / "subdir"
    subdir.mkdir()
    subsubdir = subdir / "subsubdir"
    subsubdir.mkdir()
    subdir2 = root / "subdir2"
    subdir2.mkdir()
    subfile = root / "subfile.txt"
    subfile.touch()
    filesample = subdir / "filesample.txt"
    filesample.touch()
    filesample2 = subdir / "filesample2.txt"
    filesample2.touch()

    avail_path = {
        subdir: str(swhid_of_dir(bytes(subdir))),
        subsubdir: str(swhid_of_dir(bytes(subsubdir))),
        subdir2: str(swhid_of_dir(bytes(subdir2))),
        subfile: str(swhid_of_file(bytes(subfile))),
        filesample: str(swhid_of_file(bytes(filesample))),
        filesample2: str(swhid_of_file(bytes(filesample2))),
    }

    return {
        "root": root,
        "paths": avail_path,
        "filesample": filesample,
        "filesample2": filesample2,
        "subsubdir": subsubdir,
        "subdir": subdir,
    }


@pytest.fixture(scope="function")
def example_tree(temp_folder):
    """Fixture that generate a Tree with the root present in the
       session fixture "temp_folder".
    """
    example_tree = Tree(temp_folder["root"])
    assert example_tree.path == temp_folder["root"]

    return example_tree


@pytest.fixture(scope="function")
def example_dirs(example_tree, temp_folder):
    """
        Fixture that fill the fixture example_tree with the values contained in
        the fixture temp_folder and returns the directories information of the
        filled example_tree.

    """
    root = temp_folder["root"]
    filesample_path = temp_folder["filesample"]
    filesample2_path = temp_folder["filesample2"]
    subsubdir_path = temp_folder["subsubdir"]
    known_paths = [filesample_path, filesample2_path, subsubdir_path]

    for path, swhid in temp_folder["paths"].items():
        if path in known_paths:
            example_tree.add_node(path, swhid, True)
        else:
            example_tree.add_node(path, swhid, False)

    return example_tree.get_directories_info(root)


@pytest.fixture
def test_sample_folder(datadir, tmp_path):
    """Location of the "data" folder """
    archive_path = Path(os.path.join(datadir, "sample-folder.tgz"))
    assert archive_path.exists()
    shutil.unpack_archive(archive_path, extract_dir=tmp_path)
    test_sample_folder = Path(os.path.join(tmp_path, "sample-folder"))
    assert test_sample_folder.exists()
    return test_sample_folder


@pytest.fixture
def test_swhids_sample(tmp_path):
    """Create and return the opened "swhids_sample" file,
       filled with present swhids present in data.py
    """
    test_swhids_sample = Path(os.path.join(tmp_path, "swhids_sample.txt"))

    with open(test_swhids_sample, "w") as f:
        f.write("\n".join(swhid for swhid in present_swhids))

    assert test_swhids_sample.exists()
    return open(test_swhids_sample, "r")


@pytest.fixture(scope="session")
def app():
    """Flask backend API (used by live_server)."""
    app = create_app()
    return app



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.scanner.backend import LIMIT, create_app
from swh.scanner.db import Db

from .data import present_swhids


def test_backend_endpoint_all_present(tmp_path, live_server, test_swhids_sample):
    tmp_dbfile = tmp_path / "tmp_db.sqlite"
    db = Db(tmp_dbfile)
    cur = db.conn.cursor()
    db.create_from(test_swhids_sample, LIMIT, cur)

    app = create_app(db)

    with app.test_client() as test_client:
        res = test_client.post("/api/1/known/", json=present_swhids)

        for swhid, attr in res.json.items():
            assert attr["known"]


def test_backend_endpoint_one_not_present(tmp_path, live_server, test_swhids_sample):
    tmp_dbfile = tmp_path / "tmp_db.sqlite"
    not_present_swhid = "swh:1:cnt:fa8eacf43d8646129ae8adfa1648f9307d999999"
    swhids = present_swhids + [not_present_swhid]

    db = Db(tmp_dbfile)
    cur = db.conn.cursor()
    db.create_from(test_swhids_sample, LIMIT, cur)

    app = create_app(db)

    with app.test_client() as test_client:
        res = test_client.post("/api/1/known/", json=swhids)

        for swhid, attr in res.json.items():
            if swhid != not_present_swhid:
                assert attr["known"]
            else:
                assert not attr["known"]


def test_backend_large_payload_exc(tmp_path, live_server, test_swhids_sample):
    tmp_dbfile = tmp_path / "tmp_db.sqlite"
    swhid = "swh:1:cnt:fa8eacf43d8646129ae8adfa1648f9307d999999"
    # the backend supports up to 1000 SWHID requests
    swhids = [swhid for n in range(1001)]
    db = Db(tmp_dbfile)
    cur = db.conn.cursor()
    db.create_from(test_swhids_sample, LIMIT, cur)

    app = create_app(db)

    with app.test_client() as test_client:
        res = test_client.post("/api/1/known/", json=swhids)
        assert res.status_code != 200



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from pathlib import Path
from unittest.mock import Mock, call

from click.testing import CliRunner
import pytest

import swh.scanner.cli as cli
import swh.scanner.scanner as scanner

from .data import present_swhids

DATADIR = Path(__file__).absolute().parent / "data"
CONFIG_PATH_GOOD = str(DATADIR / "global.yml")
CONFIG_PATH_GOOD2 = str(DATADIR / "global2.yml")  # alternative to global.yml
ROOTPATH_GOOD = str(DATADIR)


@pytest.fixture(scope="function")
def m_scanner(mocker):
    """Returns a mock swh.scanner.scanner object with all attributes mocked"""
    # Customizable mock of scanner module
    # Fortunately, noop is the default behavior for all methods
    scanner_mock = Mock(scanner)
    yield mocker.patch("swh.scanner.scanner", scanner_mock)


@pytest.fixture(scope="function")
def spy_configopen(mocker):
    """Returns a mock of open builtin scoped to swh.core.config"""
    yield mocker.patch("swh.core.config.open", wraps=open)


@pytest.fixture(scope="function")
def cli_runner(monkeypatch, tmp_path):
    """Return a CliRunner with default environment variable SWH_CONFIG_FILE unset"""
    BAD_CONFIG_PATH = str(tmp_path / "missing")
    monkeypatch.setattr(cli, "DEFAULT_CONFIG_PATH", BAD_CONFIG_PATH)
    return CliRunner(env={"SWH_CONFIG_FILE": None})


@pytest.fixture(scope="function")
def swhids_input_file(tmp_path):
    swhids_input_file = Path(os.path.join(tmp_path, "input_file.txt"))

    with open(swhids_input_file, "w") as f:
        f.write("\n".join(swhid for swhid in present_swhids))

    assert swhids_input_file.exists()
    return swhids_input_file


# TEST BEGIN

# For nominal code paths, check that the right config file is loaded
# scanner is mocked to not run actual scan, config loading is mocked to check its usage


def test_smoke(cli_runner):
    """Break if basic functionality breaks"""
    res = cli_runner.invoke(cli.scanner, ["scan", "-h"])
    assert res.exit_code == 0


def test_config_path_option_bad(cli_runner, tmp_path):
    """Test bad option no envvar bad default"""
    CONFPATH_BAD = str(tmp_path / "missing")
    res = cli_runner.invoke(cli.scanner, ["-C", CONFPATH_BAD, "scan", ROOTPATH_GOOD])
    assert res.exit_code != 0


def test_default_config_path(cli_runner, m_scanner, spy_configopen, monkeypatch):
    """Test no option no envvar good default"""
    monkeypatch.setattr(cli, "DEFAULT_CONFIG_PATH", CONFIG_PATH_GOOD)
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_GOOD])
    assert res.exit_code == 0
    assert spy_configopen.call_args == call(CONFIG_PATH_GOOD)
    assert m_scanner.scan.call_count == 1


def test_root_no_config(cli_runner, m_scanner, spy_configopen):
    """Test no config = no option no envvar bad default, good root"""
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_GOOD])
    assert res.exit_code == 0
    assert spy_configopen.call_count == 0
    assert m_scanner.scan.call_count == 1


def test_root_bad(cli_runner, tmp_path):
    """Test no option no envvar bad default bad root"""
    ROOTPATH_BAD = str(tmp_path / "missing")
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_BAD])
    assert res.exit_code != 0


def test_config_path_envvar_good(cli_runner, m_scanner, spy_configopen):
    """Test no option good envvar bad default good root"""
    cli_runner.env["SWH_CONFIG_FILE"] = CONFIG_PATH_GOOD
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_GOOD])
    assert res.exit_code == 0
    assert spy_configopen.call_args == call(CONFIG_PATH_GOOD)
    assert m_scanner.scan.call_count == 1


def test_config_path_envvar_bad(cli_runner, tmp_path):
    """Test no option bad envvar bad default good root"""
    CONFPATH_BAD = str(tmp_path / "missing")
    cli_runner.env["SWH_CONFIG_FILE"] = CONFPATH_BAD
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_GOOD])
    assert res.exit_code != 0


def test_config_path_option_envvar(cli_runner, m_scanner, spy_configopen):
    """Test good option good envvar bad default good root
    Check that option has precedence over envvar"""
    cli_runner.env["SWH_CONFIG_FILE"] = CONFIG_PATH_GOOD2
    res = cli_runner.invoke(
        cli.scanner, ["-C", CONFIG_PATH_GOOD, "scan", ROOTPATH_GOOD]
    )
    assert res.exit_code == 0
    assert spy_configopen.call_args == call(CONFIG_PATH_GOOD)
    assert m_scanner.scan.call_count == 1


def test_api_url_option(cli_runner, m_scanner):
    """Test no config good root good url"""
    API_URL = "https://example.com/api/1"  # without trailing "/"
    res = cli_runner.invoke(cli.scanner, ["scan", ROOTPATH_GOOD, "-u", API_URL])
    assert res.exit_code == 0
    assert m_scanner.scan.call_count == 1


def test_db_option(cli_runner, swhids_input_file, tmp_path):
    res = cli_runner.invoke(
        cli.scanner,
        [
            "db",
            "import",
            "--input",
            swhids_input_file,
            "--output",
            f"{tmp_path}/test_db.sqlite",
        ],
    )
    assert res.exit_code == 0



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from enum import Enum
import json
from pathlib import Path
import sys
from typing import Any, Dict, Iterator, List, Tuple

import ndjson

from swh.model.identifiers import CONTENT, DIRECTORY

from .exceptions import InvalidDirectoryPath, InvalidObjectType
from .plot import generate_sunburst, offline_plot


class Color(Enum):
    blue = "\033[94m"
    green = "\033[92m"
    red = "\033[91m"
    end = "\033[0m"


def colorize(text: str, color: Color):
    return color.value + text + Color.end.value


class Tree:
    """Representation of a file system structure
    """

    def __init__(self, path: Path, father: Tree = None):
        self.father = father
        self.path = path
        self.otype = DIRECTORY if path.is_dir() else CONTENT
        self.swhid = ""
        self.known = False
        self.children: Dict[Path, Tree] = {}

    def add_node(self, path: Path, swhid: str, known: bool) -> None:
        """Recursively add a new path.
        """
        relative_path = path.relative_to(self.path)

        if relative_path == Path("."):
            self.swhid = swhid
            self.known = known
            return

        new_path = self.path.joinpath(relative_path.parts[0])
        if new_path not in self.children:
            self.children[new_path] = Tree(new_path, self)

        self.children[new_path].add_node(path, swhid, known)

    def show(self, fmt) -> None:
        """Show tree in different formats"""
        if fmt == "json":
            print(json.dumps(self.to_dict(), indent=4, sort_keys=True))

        if fmt == "ndjson":
            print(
                ndjson.dumps(
                    {str(Path(k).relative_to(self.path)): v}
                    for node in self.iterate()
                    for k, v in node.attributes.items()
                )
            )

        elif fmt == "text":
            isatty = sys.stdout.isatty()
            root_dir = self.path.relative_to(self.path.parent)
            print(colorize(str(root_dir), Color.blue) if isatty else str(root_dir))
            self.print_children(isatty)

        elif fmt == "sunburst":
            root = self.path
            directories = self.get_directories_info(root)
            sunburst = generate_sunburst(directories, root)
            offline_plot(sunburst)

    def print_children(self, isatty: bool, inc: int = 1) -> None:
        for path, node in self.children.items():
            self.print_node(node, isatty, inc)
            if node.children:
                node.print_children(isatty, inc + 1)

    def print_node(self, node: Any, isatty: bool, inc: int) -> None:
        rel_path = str(node.path.relative_to(self.path))
        begin = "│   " * inc
        end = "/" if node.otype == DIRECTORY else ""

        if isatty:
            if not node.known:
                rel_path = colorize(rel_path, Color.red)
            elif node.otype == DIRECTORY:
                rel_path = colorize(rel_path, Color.blue)
            elif node.otype == CONTENT:
                rel_path = colorize(rel_path, Color.green)

        print(f"{begin}{rel_path}{end}")

    @property
    def attributes(self) -> Dict[str, Dict[str, Any]]:
        """
        Get the attributes of the current node grouped by the relative path.

        Returns:
            a dictionary containing a path as key and its known/unknown status and the
            SWHID as values.

        """
        return {str(self.path): {"swhid": self.swhid, "known": self.known,}}

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Recursively flatten the current tree nodes into a dictionary.

        For example, if you have the following structure:

        .. code-block:: none

            root {
                subdir: {
                    file.txt
                }
            }

        The generated dictionary will be:

        .. code-block:: none

            {
                "root": {
                    "swhid": "...",
                    "known": True/False
                }
                "root/subdir": {
                    "swhid": "...",
                    "known": True/False
                }
                "root/subdir/file.txt": {
                    "swhid": "...",
                    "known": True/False
                }
            }
        """
        return {
            str(Path(k).relative_to(self.path)): v
            for node in self.iterate()
            for k, v in node.attributes.items()
        }

    def iterate(self) -> Iterator[Tree]:
        """
        Recursively iterate through the children of the current node

        """
        for _, child_node in self.children.items():
            yield child_node
            if child_node.otype == DIRECTORY:
                yield from child_node.iterate()

    def get_files_from_dir(self, dir_path: Path) -> List:
        """
        Retrieve files information about a specific directory path

        Returns:
            A list containing the files attributes present inside the directory given
            in input
        """

        def get_files(node):
            files = []
            for _, node in node.children.items():
                if node.otype == CONTENT:
                    files.append(node.attributes)
            return files

        if dir_path == self.path:
            return get_files(self)
        else:
            for node in self.iterate():
                if node.path == dir_path:
                    return get_files(node)
            raise InvalidDirectoryPath(
                "The directory provided doesn't match any stored directory"
            )

    def _get_sub_dirs_info(self, root, directories):
        """Fills the directories given in input with the contents information
           stored inside the directory child, only if they have contents.
        """
        for path, child_node in self.children.items():
            if child_node.otype == DIRECTORY:
                rel_path = path.relative_to(root)
                contents_info = child_node.count_contents()
                # checks the first element of the tuple
                # (the number of contents in a directory)
                # if it is equal to zero it means that there are no contents
                # in that directory.
                if not contents_info[0] == 0:
                    directories[rel_path] = contents_info
                if child_node.has_dirs():
                    child_node._get_sub_dirs_info(root, directories)

    def get_directories_info(self, root: Path) -> Dict[Path, Tuple[int, int]]:
        """Get information about all directories under the given root.

        Returns:
            A dictionary with a directory path as key and the relative
            contents information (the result of count_contents) as values.

        """
        directories = {root: self.count_contents()}
        self._get_sub_dirs_info(root, directories)
        return directories

    def count_contents(self) -> Tuple[int, int]:
        """Count how many contents are present inside a directory.
           If a directory has a SWHID returns as it has all the contents.

        Returns:
            A tuple with the total number of the contents and the number
            of contents known (the ones that have a persistent identifier).

        """
        contents = 0
        discovered = 0

        if not self.otype == DIRECTORY:
            raise InvalidObjectType(
                "Can't count contents of the object type: %s" % self.otype
            )

        if self.known:
            # to identify a directory with all files/directories present
            return (1, 1)
        else:
            for _, child_node in self.children.items():
                if child_node.otype == CONTENT:
                    contents += 1
                    if child_node.known:
                        discovered += 1

        return (contents, discovered)

    def has_dirs(self) -> bool:
        """Checks if node has directories
        """
        for _, child_node in self.children.items():
            if child_node.otype == DIRECTORY:
                return True
        return False



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from flask import Flask, request

from .db import Db
from .exceptions import LargePayloadExc

LIMIT = 1000


def create_app(db: Db):
    """Backend for swh-scanner, implementing the /known endpoint of the
       Software Heritage Web API"""
    app = Flask(__name__)

    @app.route("/api/1/known/", methods=["POST"])
    def known():
        swhids = request.get_json()

        if len(swhids) > LIMIT:
            raise LargePayloadExc(
                f"The maximum number of SWHIDs this endpoint can receive is {LIMIT}"
            )

        cur = db.conn.cursor()
        res = {swhid: {"known": db.known(swhid, cur)} for swhid in swhids}
        cur.close()

        return res

    return app


def run(host: str, port: int, db: Db):
    """Serve the local database
    """
    app = create_app(db)
    app.run(host, port, debug=True)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module is an interface to interact with the local database
where the SWHIDs will be saved for the local API.

SWHIDs can be added directly from an input file.
"""

from io import TextIOWrapper
import logging
from pathlib import Path
import re
import sqlite3
from typing import Iterable

from swh.core.utils import grouper

from .exceptions import DBError

# XXX copied and simplified from swh.model.identifiers (WIP), replace this in favor of
# swh.model.identifiers.SWHID_RE when it is landed there
SWHID_RE = re.compile("^swh:1:(ori|snp|rel|rev|dir|cnt):[0-9a-f]{40}$")


class Db:
    """Local database interface"""

    def __init__(self, db_file: Path):
        self.db_file: Path = db_file
        self.conn: sqlite3.Connection = sqlite3.connect(
            db_file, check_same_thread=False
        )

    def close(self):
        """Close the connection to the database."""
        self.conn.close()

    def create_table(self, cur: sqlite3.Cursor):
        """Create the table where the SWHIDs will be stored."""
        cur.execute("""CREATE TABLE IF NOT EXISTS swhids (swhid text PRIMARY KEY)""")

    def add(self, swhids: Iterable[str], chunk_size: int, cur: sqlite3.Cursor):
        """Insert the SWHID inside the database."""
        for swhids_chunk in grouper(swhids, chunk_size):
            cur.executemany(
                """INSERT INTO swhids VALUES (?)""",
                [(swhid_chunk,) for swhid_chunk in swhids_chunk],
            )

    @staticmethod
    def iter_swhids(lines: Iterable[str]) -> Iterable[str]:
        lineno = 0
        for line in lines:
            lineno += 1
            swhid = line.rstrip()
            if SWHID_RE.match(swhid):
                yield swhid
            else:
                logging.error("ignoring invalid SWHID on line %d: %s", lineno, swhid)

    def create_from(
        self, input_file: TextIOWrapper, chunk_size: int, cur: sqlite3.Cursor
    ):
        """Create a new database with the SWHIDs present inside the input file."""
        try:
            self.create_table(cur)
            cur.execute("PRAGMA synchronous = OFF")
            cur.execute("PRAGMA journal_mode = OFF")
            self.add(self.iter_swhids(input_file), chunk_size, cur)
            cur.close()
            self.conn.commit()
        except sqlite3.Error as e:
            raise DBError(f"SQLite error: {e}")

    def known(self, swhid: str, cur: sqlite3.Cursor):
        """Check if a given SWHID is present or not inside the local database."""
        cur.execute("""SELECT 1 FROM swhids WHERE swhid=?""", (swhid,))
        res = cur.fetchone()

        return res is not None



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class InvalidObjectType(TypeError):
    pass


class InvalidDirectoryPath(Exception):
    pass


class LargePayloadExc(Exception):
    pass


class DBError(Exception):
    pass


class APIError(Exception):
    def __str__(self):
        return '"%s"' % self.args


def error_response(reason: str, status_code: int, api_url: str):
    error_msg = f"{status_code} {reason}: '{api_url}'"
    raise APIError(error_msg)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import os
from typing import Any, Dict, Optional

import click
from importlib_metadata import version
import yaml

from swh.core import config
from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

from .exceptions import DBError

# Config for the "serve" option
BACKEND_DEFAULT_PORT = 5011

# All generic config code should reside in swh.core.config
CONFIG_ENVVAR = "SWH_CONFIG_FILE"
DEFAULT_CONFIG_PATH = os.path.join(click.get_app_dir("swh"), "global.yml")

DEFAULT_CONFIG: Dict[str, Any] = {
    "web-api": {
        "url": "https://archive.softwareheritage.org/api/1/",
        "auth-token": None,
    }
}


CONFIG_FILE_HELP = f"""Configuration file:

\b
The CLI option or the environment variable will fail if invalid.
CLI option is checked first.
Then, environment variable {CONFIG_ENVVAR} is checked.
Then, if cannot load the default path, a set of default values are used.
Default config path is {DEFAULT_CONFIG_PATH}.
Default config values are:

\b
{yaml.dump(DEFAULT_CONFIG)}"""
SCANNER_HELP = f"""Software Heritage Scanner tools.

{CONFIG_FILE_HELP}"""


def setup_config(ctx, api_url):
    config = ctx.obj["config"]
    if api_url:
        if not api_url.endswith("/"):
            api_url += "/"
        config["web-api"]["url"] = api_url

    return config


@swh_cli_group.group(
    name="scanner", context_settings=CONTEXT_SETTINGS, help=SCANNER_HELP,
)
@click.option(
    "-C",
    "--config-file",
    default=None,
    type=click.Path(exists=False, dir_okay=False, path_type=str),
    help="""YAML configuration file""",
)
@click.version_option(
    version=version("swh-scanner"), prog_name="swh-scanner",
)
@click.pass_context
def scanner(ctx, config_file: Optional[str]):

    env_config_path = os.environ.get(CONFIG_ENVVAR)

    # read_raw_config do not fail if file does not exist, so check it beforehand
    # while enforcing loading priority
    if config_file:
        if not config.config_exists(config_file):
            raise click.BadParameter(
                f"File '{config_file}' cannot be opened.", param_hint="--config-file"
            )
    elif env_config_path:
        if not config.config_exists(env_config_path):
            raise click.BadParameter(
                f"File '{env_config_path}' cannot be opened.", param_hint=CONFIG_ENVVAR
            )
        config_file = env_config_path
    elif config.config_exists(DEFAULT_CONFIG_PATH):
        config_file = DEFAULT_CONFIG_PATH

    conf = DEFAULT_CONFIG
    if config_file is not None:
        conf = config.read_raw_config(config.config_basepath(config_file))
        conf = config.merge_configs(DEFAULT_CONFIG, conf)

    ctx.ensure_object(dict)
    ctx.obj["config"] = conf


@scanner.command(name="scan")
@click.argument("root_path", required=True, type=click.Path(exists=True))
@click.option(
    "-u",
    "--api-url",
    default=None,
    metavar="API_URL",
    show_default=True,
    help="URL for the api request",
)
@click.option(
    "--exclude",
    "-x",
    "patterns",
    metavar="PATTERN",
    multiple=True,
    help="Exclude directories using glob patterns \
    (e.g., ``*.git`` to exclude all .git directories)",
)
@click.option(
    "-f",
    "--output-format",
    "out_fmt",
    default="text",
    show_default=True,
    type=click.Choice(["text", "json", "ndjson", "sunburst"], case_sensitive=False),
    help="The output format",
)
@click.option(
    "-i", "--interactive", is_flag=True, help="Show the result in a dashboard"
)
@click.pass_context
def scan(ctx, root_path, api_url, patterns, out_fmt, interactive):
    """Scan a source code project to discover files and directories already
    present in the archive"""
    import swh.scanner.scanner as scanner

    config = setup_config(ctx, api_url)
    scanner.scan(config, root_path, patterns, out_fmt, interactive)


@scanner.group("db", help="Manage local knowledge base for swh-scanner")
@click.pass_context
def db(ctx):
    pass


@db.command("import")
@click.option(
    "-i",
    "--input",
    "input_file",
    metavar="INPUT_FILE",
    required=True,
    type=click.File("r"),
    help="A file containing SWHIDs",
)
@click.option(
    "-o",
    "--output",
    "output_file_db",
    metavar="OUTPUT_DB_FILE",
    required=True,
    show_default=True,
    help="The name of the generated sqlite database",
)
@click.option(
    "-s",
    "--chunk-size",
    "chunk_size",
    default="10000",
    metavar="SIZE",
    show_default=True,
    type=int,
    help="The chunk size ",
)
@click.pass_context
def import_(ctx, chunk_size, input_file, output_file_db):
    """Create SQLite database of known SWHIDs from a textual list of SWHIDs

    """
    from .db import Db

    db = Db(output_file_db)
    cur = db.conn.cursor()
    try:
        db.create_from(input_file, chunk_size, cur)
        db.close()
    except DBError as e:
        ctx.fail("Failed to import SWHIDs into database: {0}".format(e))


@db.command("serve")
@click.option(
    "-h",
    "--host",
    metavar="HOST",
    default="127.0.0.1",
    show_default=True,
    help="The host of the API server",
)
@click.option(
    "-p",
    "--port",
    metavar="PORT",
    default=f"{BACKEND_DEFAULT_PORT}",
    show_default=True,
    help="The port of the API server",
)
@click.option(
    "-f",
    "--db-file",
    "db_file",
    metavar="DB_FILE",
    default="SWHID_DB.sqlite",
    show_default=True,
    type=click.Path(exists=True),
    help="An sqlite database file (it can be generated with: 'swh scanner db import')",
)
@click.pass_context
def serve(ctx, host, port, db_file):
    """Start an API service using the sqlite database generated with the "db import"
    option."""
    import swh.scanner.backend as backend

    from .db import Db

    db = Db(db_file)
    backend.run(host, port, db)
    db.close()


def main():
    return scanner(auto_envvar_prefix="SWH_SCANNER")


if __name__ == "__main__":
    main()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import itertools
import os
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Pattern, Tuple, Union

import aiohttp

from swh.model.from_disk import (
    Content,
    Directory,
    accept_all_directories,
    extract_regex_objs,
)
from swh.model.identifiers import CoreSWHID, ObjectType

from .dashboard.dashboard import run_app
from .exceptions import error_response
from .model import Tree
from .plot import generate_sunburst


async def swhids_discovery(
    swhids: List[str], session: aiohttp.ClientSession, api_url: str,
) -> Dict[str, Dict[str, bool]]:
    """API Request to get information about the SoftWare Heritage persistent
    IDentifiers (SWHIDs) given in input.

    Args:
        swhids: a list of SWHIDS
        api_url: url for the API request

    Returns:
        A dictionary with:

        key:
            SWHID searched
        value:
            value['known'] = True if the SWHID is found
            value['known'] = False if the SWHID is not found

    """
    endpoint = api_url + "known/"
    chunk_size = 1000
    requests = []

    def get_chunk(swhids):
        for i in range(0, len(swhids), chunk_size):
            yield swhids[i : i + chunk_size]

    async def make_request(swhids):
        async with session.post(endpoint, json=swhids) as resp:
            if resp.status != 200:
                error_response(resp.reason, resp.status, endpoint)

            return await resp.json()

    if len(swhids) > chunk_size:
        for swhids_chunk in get_chunk(swhids):
            requests.append(asyncio.create_task(make_request(swhids_chunk)))

        res = await asyncio.gather(*requests)
        # concatenate list of dictionaries
        return dict(itertools.chain.from_iterable(e.items() for e in res))
    else:
        return await make_request(swhids)


def directory_filter(
    path_name: Union[str, bytes], exclude_patterns: Iterable[Pattern[bytes]]
) -> bool:
    """It checks if the path_name is matching with the patterns given in input.

    It is also used as a `dir_filter` function when generating the directory
    object from `swh.model.from_disk`

    Returns:
        False if the directory has to be ignored, True otherwise

    """
    path = Path(path_name.decode() if isinstance(path_name, bytes) else path_name)

    for sre_pattern in exclude_patterns:
        if sre_pattern.match(bytes(path)):
            return False
    return True


def get_subpaths(
    path: Path, exclude_patterns: Iterable[Pattern[bytes]]
) -> Iterator[Tuple[Path, str]]:
    """Find the SoftWare Heritage persistent IDentifier (SWHID) of
    the directories and files under a given path.

    Args:
        path: the root path

    Yields:
        pairs of: path, the relative SWHID

    """

    def swhid_of(path: Path) -> str:
        if path.is_dir():
            if exclude_patterns:

                def dir_filter(dirpath: bytes, *args) -> bool:
                    return directory_filter(dirpath, exclude_patterns)

            else:
                dir_filter = accept_all_directories  # type: ignore

            obj = Directory.from_disk(
                path=bytes(path), dir_filter=dir_filter
            ).get_data()

            return str(CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=obj["id"]))
        else:
            obj = Content.from_file(path=bytes(path)).get_data()
            return str(
                CoreSWHID(object_type=ObjectType.CONTENT, object_id=obj["sha1_git"])
            )

    dirpath, dnames, fnames = next(os.walk(path))
    for node in itertools.chain(dnames, fnames):
        sub_path = Path(dirpath).joinpath(node)
        yield (sub_path, swhid_of(sub_path))


async def parse_path(
    path: Path,
    session: aiohttp.ClientSession,
    api_url: str,
    exclude_patterns: Iterable[Pattern[bytes]],
) -> Iterator[Tuple[str, str, bool]]:
    """Check if the sub paths of the given path are present in the
    archive or not.

    Args:
        path: the source path
        api_url: url for the API request

    Returns:
        a map containing tuples with: a subpath of the given path,
        the SWHID of the subpath and the result of the api call

    """
    parsed_paths = dict(get_subpaths(path, exclude_patterns))
    parsed_swhids = await swhids_discovery(
        list(parsed_paths.values()), session, api_url
    )

    def unpack(tup):
        subpath, swhid = tup
        return (subpath, swhid, parsed_swhids[swhid]["known"])

    return map(unpack, parsed_paths.items())


async def run(
    config: Dict[str, Any],
    root: str,
    source_tree: Tree,
    exclude_patterns: Iterable[Pattern[bytes]],
) -> None:
    """Start scanning from the given root.

    It fills the source tree with the path discovered.

    Args:
        root: the root path to scan
        api_url: url for the API request

    """
    api_url = config["web-api"]["url"]

    async def _scan(root, session, api_url, source_tree, exclude_patterns):
        for path, obj_swhid, known in await parse_path(
            root, session, api_url, exclude_patterns
        ):
            obj_type = CoreSWHID.from_string(obj_swhid).object_type

            if obj_type == ObjectType.CONTENT:
                source_tree.add_node(path, obj_swhid, known)
            elif obj_type == ObjectType.DIRECTORY and directory_filter(
                path, exclude_patterns
            ):
                source_tree.add_node(path, obj_swhid, known)
                if not known:
                    await _scan(path, session, api_url, source_tree, exclude_patterns)

    if config["web-api"]["auth-token"]:
        headers = {"Authorization": f"Bearer {config['web-api']['auth-token']}"}
    else:
        headers = {}

    async with aiohttp.ClientSession(headers=headers, trust_env=True) as session:
        await _scan(root, session, api_url, source_tree, exclude_patterns)


def scan(
    config: Dict[str, Any],
    root_path: str,
    exclude_patterns: Iterable[str],
    out_fmt: str,
    interactive: bool,
):
    """Scan a source code project to discover files and directories already
    present in the archive"""
    converted_patterns = set(pattern.encode() for pattern in exclude_patterns)
    sre_patterns = set()
    if exclude_patterns:
        sre_patterns = {
            reg_obj
            for reg_obj in extract_regex_objs(root_path.encode(), converted_patterns)
        }

    source_tree = Tree(Path(root_path))
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(config, root_path, source_tree, sre_patterns))

    if interactive:
        root = Path(root_path)
        directories = source_tree.get_directories_info(root)
        figure = generate_sunburst(directories, root)
        run_app(figure, source_tree)
    else:
        source_tree.show(out_fmt)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from flask import url_for
import pytest

from swh.model.exceptions import InvalidDirectoryPath
from swh.scanner.exceptions import APIError
from swh.scanner.model import Tree
from swh.scanner.scanner import extract_regex_objs, get_subpaths, run, swhids_discovery

from .data import correct_api_response, present_swhids, to_exclude_swhid

aio_url = "http://example.org/api/known/"


def test_extract_regex_objs(temp_folder):
    root_path = bytes(temp_folder["root"])

    patterns = (bytes(temp_folder["subdir"]), b"/none")

    sre_patterns = [reg_obj for reg_obj in extract_regex_objs(root_path, patterns)]
    assert len(sre_patterns) == 2

    patterns = (*patterns, b"/tmp")
    with pytest.raises(InvalidDirectoryPath):
        sre_patterns = [reg_obj for reg_obj in extract_regex_objs(root_path, patterns)]


def test_scanner_correct_api_request(mock_aioresponse, event_loop, aiosession):
    mock_aioresponse.post(
        aio_url,
        status=200,
        content_type="application/json",
        body=json.dumps(correct_api_response),
    )

    actual_result = event_loop.run_until_complete(
        swhids_discovery([], aiosession, "http://example.org/api/")
    )

    assert correct_api_response == actual_result


def test_scanner_raise_apierror(mock_aioresponse, event_loop, aiosession):
    mock_aioresponse.post(aio_url, content_type="application/json", status=413)

    with pytest.raises(APIError):
        event_loop.run_until_complete(
            swhids_discovery([], aiosession, "http://example.org/api/")
        )


def test_scanner_raise_apierror_input_size_limit(event_loop, aiosession, live_server):

    api_url = url_for("index", _external=True)
    request = [
        "swh:1:cnt:7c4c57ba9ff496ad179b8f65b1d286edbda34c9a" for i in range(901)
    ]  # /known/ is limited at 900

    with pytest.raises(APIError):
        event_loop.run_until_complete(swhids_discovery(request, aiosession, api_url))


def test_scanner_get_subpaths(temp_folder):
    root = temp_folder["root"]

    actual_result = []
    for subpath, swhid in get_subpaths(root, tuple()):
        # also check if it's a symlink since pytest tmp_dir fixture create
        # also a symlink to each directory inside the tmp_dir path
        if subpath.is_dir() and not subpath.is_symlink():
            actual_result.append((subpath, swhid))

    assert len(actual_result) == 2


@pytest.mark.options(debug=False)
def test_app(app):
    assert not app.debug


def test_scanner_result(live_server, event_loop, test_sample_folder):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}

    source_tree = Tree(test_sample_folder)
    event_loop.run_until_complete(run(config, test_sample_folder, source_tree, set()))

    for child_node in source_tree.iterate():
        node_info = list(child_node.attributes.values())[0]
        if node_info["swhid"] in present_swhids:
            assert node_info["known"] is True
        else:
            assert node_info["known"] is False


def test_scanner_result_with_exclude_patterns(
    live_server, event_loop, test_sample_folder
):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}
    to_exclude_dir = str(test_sample_folder) + "/toexclude"

    patterns = (to_exclude_dir.encode(),)
    exclude_pattern = {
        reg_obj for reg_obj in extract_regex_objs(bytes(test_sample_folder), patterns)
    }

    source_tree = Tree(test_sample_folder)
    event_loop.run_until_complete(
        run(config, test_sample_folder, source_tree, exclude_pattern)
    )

    for child_node in source_tree.iterate():
        node_info = list(child_node.attributes.values())[0]
        assert node_info["swhid"] != to_exclude_swhid



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.scanner.plot import (
    build_hierarchical_df,
    compute_max_depth,
    generate_df_from_dirs,
)


def test_max_depth(temp_folder, example_dirs):
    root = temp_folder["root"]
    max_depth = compute_max_depth(example_dirs, root)
    assert max_depth == 2


def test_generate_df_from_dirs(temp_folder, example_dirs):
    root = temp_folder["root"]
    max_depth = compute_max_depth(example_dirs, root)
    metrics_columns = ["contents", "known"]
    levels_columns = ["lev" + str(i) for i in range(max_depth)]
    df_columns = levels_columns + metrics_columns

    actual_df = generate_df_from_dirs(example_dirs, df_columns, root, max_depth)

    # assert root is empty
    assert actual_df["lev0"][0] == ""
    assert actual_df["lev1"][0] == ""

    # assert subdir has correct contents information
    assert actual_df["contents"][1] == 2
    assert actual_df["known"][1] == 2

    # assert subsubdir has correct level information
    assert actual_df["lev0"][2] == "subdir"
    assert actual_df["lev1"][2] == "subdir/subsubdir"


def test_build_hierarchical_df(temp_folder, example_dirs):
    root = temp_folder["root"]
    max_depth = compute_max_depth(example_dirs, root)
    metrics_columns = ["contents", "known"]
    levels_columns = ["lev" + str(i) for i in range(max_depth)]
    df_columns = levels_columns + metrics_columns

    actual_df = generate_df_from_dirs(example_dirs, df_columns, root, max_depth)

    actual_result = build_hierarchical_df(
        actual_df, levels_columns, metrics_columns, root
    )

    assert actual_result["parent"][1] == "subdir"
    assert actual_result["contents"][1] == 2
    assert actual_result["id"][5] == root
    assert actual_result["known"][5] == 75
